from threading import Event
from typing import Iterator
from rich.console import Console

from rich.progress import Progress, TextColumn, TimeElapsedColumn
from rich.live import Live
from rich.panel import Panel

from bq_meta import const


def progress(console: Console, name: str, iterator: Iterator) -> list:
    size = 0
    result = []
    with Progress(
        TextColumn("[progress.description]{task.description}", style=const.info_style),
        "•",
        TextColumn("[progress.completed]{task.completed}"),
        "•",
        TimeElapsedColumn(),
        console=console,
        transient=True,
    ) as progress:
        task = progress.add_task(f"Fetching {name}")
        for item in iterator:
            result.append(item)
            progress.advance(task)
            size += 1
    return result


def flash_panel(live: Live, panel: Panel):
    event = Event()
    panel.border_style = const.request_style  # border will flash a short period of time
    live.update(panel, refresh=True)
    event.wait(0.1)
    if not event.is_set():
        panel.border_style = const.border_style
        live.update(panel, refresh=True)



from google.cloud import bigquery
from rich.console import Group
from rich.table import Table
from rich.rule import Rule
from rich.text import Text
from rich.align import Align

from bq_meta import const
from bq_meta.config import Config
from bq_meta.util.num_utils import bytes_fmt, num_fmt

header = """
██▄ █ ▄▀  ▄▀▄ █ █ ██▀ █▀▄ ▀▄▀   █▄ ▄█ ██▀ ▀█▀ ▄▀▄ █▀▄ ▄▀▄ ▀█▀ ▄▀▄
█▄█ █ ▀▄█ ▀▄█ ▀▄█ █▄▄ █▀▄  █    █ ▀ █ █▄▄  █  █▀█ █▄▀ █▀█  █  █▀█
"""

header_renderable = Align(
    header,
    align="center",
    style=const.info_style,
)


def get_config_info(config: Config) -> Group:
    return Group(text_tuple("Account", config.account))


# fmt: off
def get_table_output(table: bigquery.Table) -> Group:
    size = bytes_fmt(table.num_bytes)
    long_term_size = bytes_fmt(int(table._properties.get("numLongTermBytes", 0)))
    rows = num_fmt(table.num_rows)
    
    created = table.created.strftime("%Y-%m-%d %H:%M:%S UTC")
    modified = table.modified.strftime("%Y-%m-%d %H:%M:%S UTC")
    expiry = table.expires.strftime("%Y-%m-%d %H:%M:%S UTC") if table.expires else None
    
    partitioned_by = table.time_partitioning.type_ if table.time_partitioning else None
    partitioned_field = table.time_partitioning.field if table.time_partitioning else None
    partition_filter = table.require_partition_filter if table.require_partition_filter else None
    _num_partitions = table._properties.get("numPartitions", None)
    num_of_partitions = int(_num_partitions) if _num_partitions else None
    
    streaming_buffer_size = bytes_fmt(table.streaming_buffer.estimated_bytes) if table.streaming_buffer else None
    streaming_buffer_rows = num_fmt(table.streaming_buffer.estimated_rows) if table.streaming_buffer else None
    streaming_entry_time = table.streaming_buffer.oldest_entry_time.strftime("%Y-%m-%d %H:%M:%S UTC") if table.streaming_buffer else None
    return Group(
        Rule(style=const.darker_style),
        text_tuple("Table ID", table.full_table_id),
        text_tuple("Description", table.description),
        text_tuple("Data location", table.location),
        Rule(style=const.darker_style),
        text_tuple("Table size", size),
        text_tuple("Long-term size", long_term_size),
        text_tuple("Number of rows", rows),
        Rule(style=const.darker_style),
        table_tuple({
            "Created": created, 
            "Last modified": modified, 
            "Table expiry": expiry 
        }),
        Rule(style=const.darker_style),
        text_tuple("Partitioned by", partitioned_by),
        text_tuple("Partitioned on field", partitioned_field),
        text_tuple("Partition filter", partition_filter),
        text_tuple("Partitions number", num_of_partitions),
        Rule(style=const.darker_style),
        text_tuple("Clustered by", table.clustering_fields),
        Rule(style=const.darker_style),
        text_tuple("Streaming buffer rows", streaming_buffer_rows),
        text_tuple("Streaming buffer size", streaming_buffer_size),
        text_tuple("Streaming entry time", streaming_entry_time),
        Rule(style=const.darker_style),
    )
# fmt: on


def table_tuple(tuples: dict) -> Text:
    table = Table(
        box=const.equal_box, show_header=False, show_edge=False, pad_edge=False, border_style=const.darker_style
    )
    for key, value in tuples.items():
        text = value if isinstance(tuple, Text) else Text(str(value), style="default")
        table.add_row(Text(key, style=const.key_style), text)
    return table


def text_tuple(name: str, value) -> Text:
    text = None
    if isinstance(value, Text):
        text = value
    else:
        text = Text(str(value), style="default")
    return Text(name, style=const.key_style).append(" = ", style=const.darker_style).append(text)



import webbrowser
from datetime import datetime
from typing import Optional

import readchar
from bq_meta import const, output
from bq_meta.service.history_service import HistoryService
from bq_meta.service.table_service import TableService
from google.cloud import bigquery
from rich.console import Console
from rich.live import Live
from rich.panel import Panel
from rich.layout import Layout
from rich.console import Group, RenderableType

from bq_meta.util import table_utils
from bq_meta.util.rich_utils import flash_panel


class Window:
    def __init__(
        self,
        console: Console,
        history_service: HistoryService,
        table_service: TableService,
    ):
        self.project_id: Optional[str] = None
        self.dataset_id: Optional[str] = None
        self.table_id: Optional[str] = None
        self.table: Optional[bigquery.Table] = None
        self.panel: Optional[Panel] = None
        self.subtitle = "open (o) | history (h) | quit (q)"
        self.console = console
        self.history_service = history_service
        self.table_service = table_service

    def live_window(self, table: Optional[bigquery.Table]):
        self.table = table
        with Live(self.panel, auto_refresh=False, screen=True, transient=True) as live:
            if self.table:
                table_content = output.get_table_output(self.table)
                self._update_panel(live, table_content)
            else:
                self._update_panel(live, None)
            self._loop(live)

    def _update_panel(self, live: Live, content: Optional[RenderableType]) -> None:
        group = Group(output.header_renderable)
        if content:
            group = Group(output.header_renderable, content)
        now = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        self.panel = Panel(
            title=now,
            title_align="right",
            subtitle=self.subtitle,
            renderable=group,
            border_style=const.border_style,
        )
        live.update(Layout(self.panel), refresh=True)

    def _update_table_refs(self):
        """
        Update table identifiers to the state, used dataset (d) and table (t) listing
        """
        if self.table:
            self.project_id = self.table.project
            self.dataset_id = self.table.dataset_id
            self.table_id = self.table.table_id

    def _loop(self, live: Live):
        char = readchar.readkey()

        # Open new table
        if char == "o":
            self.table = self.table_service.get_table(live=live)
            if self.table:
                self.subtitle = "open (o) | refresh (r) | schema (s) | console (c) | history (h) | quit (q)"
                self.history_service.save_table(self.table)
            self._update_table_refs()
            table_content = output.get_table_output(self.table) if self.table else None
            self._update_panel(live, table_content)
            self._loop(live)

        # Refresh table
        elif char == "r" and self.table:
            self.table = self.table_service.get_fresh_table(self.table)
            table_content = output.get_table_output(self.table)
            self._update_panel(live, table_content)
            flash_panel(live, self.panel)
            self._loop(live)

        # Show schema
        elif char == "s":
            if self.table:
                self.subtitle = "open (o) | refresh (r) | table (t) | console (c) | history (h) | quit (q)"
            schema_content = table_utils.get_schema(self.table) if self.table else None
            self._update_panel(live, schema_content)
            self._loop(live)

        # Open table in the google console
        elif char == "c":
            if self.table:
                url = f"https://console.cloud.google.com/bigquery?&ws=!1m5!1m4!4m3!1s{self.table.project}!2s{self.table.dataset_id}!3s{self.table.table_id}"
                webbrowser.open(url)
            self._loop(live)

        # Show history
        elif char == "h":
            self.table = self.history_service.pick_table(live)
            if self.table:
                self.subtitle = "open (o) | refresh (r) | schema (s) | console (c) | history (h) | quit (q)"
                self.history_service.save_table(self.table)
            else:
                self.subtitle = "open (o) | history (h) | quit (q)"
            self._update_table_refs()
            table_content = output.get_table_output(self.table) if self.table else None
            self._update_panel(live, table_content)
            self._loop(live)

        # Show table
        elif char == "t":
            if self.table:
                self.subtitle = "open (o) | refresh (r) | schema (s) | console (c) | history (h) | quit (q)"
                table_content = output.get_table_output(self.table)
            self._update_panel(live, table_content)
            self._loop(live)

        # Quit program
        elif char == "q":
            live.stop()

        # Infinite loop, until stopped
        else:
            self._loop(live)



import json
from typing import List

from bq_meta import const
from google.cloud import bigquery
from google.cloud.bigquery.schema import SchemaField
from rich import box
from rich.columns import Columns
from rich.table import Table
from rich.text import Text
from rich.tree import Tree


def get_schema(table: bigquery.Table):
    schema = table.schema
    tree = Tree(Text("Schema", const.key_style))
    table = Table(box=box.SIMPLE, show_header=False)
    _scheme_tree(schema, tree)
    _scheme_table(schema, table)
    return Columns([tree, table])


def _scheme_table(schema: List[SchemaField], table: Table):
    for field in schema:
        table.add_row(field.field_type, field.mode)
        if field.field_type == "RECORD":
            _scheme_table(field.fields, table)
    return


def _scheme_tree(fields: List[SchemaField], tree: Tree):
    for field in fields:
        node = tree.add(Text(f"{field.name}"))
        if field.field_type == "RECORD":
            _scheme_tree(field.fields, node)
    return


def get_properties(table: bigquery.Table) -> str:
    return json.dumps(table._properties, indent=2)
